# MIT License

# Copyright (c) 2021 xp

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

class ErrorType:
    SYNTAX = "SyntaxError"
    CRASH = "UnexpectedPythonError"

class LPV_Exception(Exception):
    
    def __init__(self, **kwargs):
        """
        Kwargs:
            error: str
            msg: str
            line: int
            col: int | optional
            source: str
            pointer_width: int | optional
            when: str | optional
        """
        self.error:str = kwargs.pop("error")
        self.msg:str = kwargs.pop("msg")
        self.line:int = kwargs.pop("line")
        self.col:int = kwargs.pop("col") if "col" in kwargs else None
        self.source:str = kwargs.pop("source")
        self.pointer_width:int = kwargs.pop("pointer_width") if "pointer_width" in kwargs else 1
        self.when:str = kwargs.pop("when") if "when" in kwargs else None
        for attr, value in kwargs.items():
            setattr(self, attr, value)
        super().__init__(self.__repr__())
        
    def __repr__(self):
        ln = self.line+1
        s_ln = len(str(ln))
        ln_str = self.source.splitlines()[self.line]
        string = "An Error Has Occurred:" if self.when is None else f"{self.when}::Error Occurred:"
        string += "\n  "+("─"*s_ln)+"──┬─"+("─"*len(ln_str))+"─"
        string += "\n   {ln} │ {ln_str} ".format(
            ln=ln, ln_str=ln_str,
        )
        col = "\n"
        if self.col is not None:
            col = "\n   {ln_len} │ {space}{pointer}─┤col:{col}│".format(
                ln_len=" "*s_ln, space=" "*self.col,
                col=self.col+1,
                pointer="└"+("┴"*(self.pointer_width-1)),
            )
        string += col
        return string+f"\n{self.error}: {self.msg}"


# MIT License

# Copyright (c) 2021 xp

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from typing import Callable
from LPV.src.error import ErrorType, LPV_Exception
from LPV.tools.prettier import NodePrettier
import traceback

class Node:
    
    def __init__(self, line:int, col:int=None):
        self.line, self.col = line, col
        self.node_name = type(self).__name__
    def __repr__(self) -> str:
        d = self.__dict__.copy()
        d.pop("line")
        d.pop("col")
        d.pop("node_name")
        return NodePrettier(
            self.node_name,
            **d
        ).__repr__()

class NodeVisitor:
    last_node = None
    def visit(self, node: Node):
        self.last_node = node
        return getattr(self, "visit_"+node.node_name, self.visit_error)(node)
    
    def visit_error(self, node: Node):
        raise NameError(f"no visit_{node.node_name} found")
    
    def visits(self, *nodes: Node):
        return (self.visit(node) for node in nodes)
    
    def throw_error(self, error, msg, node: Node, pointer_width:int=1, **kwargs):
        raise LPV_Exception(
            error=error, msg=msg,
            line=node.line, col=node.col,
            source=self.source,
            pointer_width=pointer_width,
            when="Runtime",
            **kwargs
        )
    
    def run(self, node_tree: Node, source: str, crash_handler:Callable=print):
        self.tree = node_tree
        self.source = source
        try:
            r = self.visit(self.tree)
        except LPV_Exception as e:
            raise e
        except Exception as e:
            if crash_handler is print:
                crash_handler(traceback.format_exc())
            else:
                crash_handler(e)
            self.throw_error(
                ErrorType.CRASH,
                f"Uh oh. Looks like the interpreter got an Python Error.",
                self.last_node
            )
        else:
            return r


# MIT License

# Copyright (c) 2021 xp

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from LPV.src.token import Token, TokenTree
from LPV.src.error import ErrorType, LPV_Exception
from typing import Callable, Union
import traceback

class LPV_Lexer:
    initiliazing = False
    def __init__(self, do_count_char:bool=False):
        self.can_count = do_count_char
        if do_count_char:
            self.count_char = 0
        if not "rules" in self.__dict__:
            raise TypeError("no self.rules found")
        r = []
        for k, v in (self.rules.items()) if isinstance(self.rules, dict) else (self.rules):
            is_callable = isinstance(v, Callable)
            if not is_callable and isinstance(v, str):
                r.append((k, getattr(self, ('lex_'+v) if not v.startswith("lex_") else v)))
            elif is_callable:
                if not v.__name__.startswith("lex_"):
                    raise TypeError(
                        f"self.rules rule function must start with 'lex_': {v.__name__}"
                    )
                r.append((k, v))
            else:
                raise TypeError(
                    f"{v} is not Callable or str"
                )
        if not r:
            raise TypeError(
                "self.rules can't be empty"
            )
        self.rules = tuple(r)
        self.name = type(self).__name__
        self.initiliazing = True
    
    def init(self, source: str):
        self.source, self.len_s = source, len(source)
        self.pos, self.char = -1, None
        self.chars = ""
        self.line, self.col = 0, -1
        self.next_char()
    
    def throw_error(self, error, msg, pointer_width:int=1, **kwargs):
        raise LPV_Exception(
            error=error, msg=msg,
            line=self.line, col=self.col,
            source=self.source, when="Lexer",
            pointer_width=pointer_width,
            **kwargs
        )
    
    def increase_count(self, num:int=1):
        if self.can_count is True:
            self.count_char += num
    def reset_count(self):
        if self.can_count is True:
            self.count_char = 0
    def get_count(self):
        if self.can_count is True:
            return (self.count_char,)
        return ()
    
    def next_char(self):
        self.pos += 1
        self.col += 1
        if self.pos >= self.len_s:
            self.char = None
        else:
            self.char = self.source[self.pos]
            if self.char == "\n":
                self.line += 1
                self.col = -1
    
    def peek(self, step=1):
        pos = self.pos+step
        if pos >= self.len_s:
            return None
        return self.source[pos]
    
    def enter(self, step=1):
        for _ in range(step):
            self.chars += self.char
            self.next_char()
    
    def skip(self, step=1):
        for _ in range(step):
            self.next_char()
    
    def clear(self):
        c = self.chars
        self.chars = ""
        return c
    
    def enter_clear(self, step=1):
        self.enter(step=step)
        return self.clear()
    
    def skip_clear(self, step=1):
        self.skip(step=step)
        return self.clear()
    
    def enter_while(self, expr: Callable):
        while not self.is_eof() and expr(self.char):
            self.enter()

    def skip_while(self, expr: Callable):
        while not self.is_eof() and expr(self.char):
            self.skip()    
    
    def slice(self, step:int, start:int=0):
        r = ""
        for i in range(step):
            c = self.peek(start+i)
            if c is not None:
                r += c
        return r
    
    def match_string(self, string: str, start:int=0, incase_sensitive:bool=False):
        slc = self.slice(len(string), start)
        r = (string == slc) if incase_sensitive is False else (string.lower() == slc.lower())
        if r is True:
            self.increase_count(len(string))
        return r
    
    def match_optional(self, *objs, **kwargs):
        for o in objs:
            if self.match(o, **kwargs):
                return True
        return False
    
    def match_optional(self, *objs, **kwargs):
        for o in objs:
            if self.match(o, **kwargs):
                return True
        return False

    def match_optional_incase(self, *objs, **kwargs):
        kwargs["incase_sensitive"] = True
        for o in objs:
            if self.match(o, **kwargs):
                return True
        return False

    def match_forward(self, *objs):
        for i, o in enumerate(objs):
            if not self.match(o, start=i):
                return False
        return True
    
    def match(self, obj: Union[tuple, list, set, str, Callable], **kwargs):
        if isinstance(obj, str):
            return self.match_string(obj, **kwargs)
        elif isinstance(obj, tuple):
            return self.match_forward(*obj)
        elif isinstance(obj, list):
            return self.match_optional(*obj, **kwargs)
        elif isinstance(obj, set):
            return self.match_optional_incase(*obj, **kwargs)
        elif isinstance(obj, Callable):
            arg = self.peek(kwargs["start"]) if "start" in kwargs else self.char
            if arg is None:
                return False
            r = True if obj(arg) else False
            if r is True:
                self.increase_count()
            return r
        else:
            raise TypeError(f"Only accept type tuple, list, str or Callable: {obj}")
    
    def is_eof(self):
        return self.char is None
    
    def put(self, string: str):
        self.chars += string
    
    def lex(self, source: str, ignore_un_update_pos:bool=False, crash_handler:Callable=print):
        if self.initiliazing is False:
            raise TypeError(
                f"{type(self).__name__} is not fully initiliaze. Forgot super().__init__() in the __init__?"
            )
        self.init(source)
        tree = []
        last_items = self.rules[-1]
        try:
            while not self.is_eof():
                lc = self.line, self.col
                for rule, func in self.rules:
                    self.reset_count()
                    if self.match(rule):
                        if ignore_un_update_pos is False:
                            p = self.pos
                            r = func(*self.get_count())
                            if self.pos == p:
                                raise TypeError(
                                    f"{func.__name__} must move to the next position atleast once"
                                    "(avoiding infinity loop)"
                                )
                        else:
                            r = func()
                        if r is None:
                            break
                        if not isinstance(r, Token):
                            raise TypeError("function must return Token object")
                        if None in (r.line, r.col):
                            r.set_lc(*lc)
                        tree.append(r)
                        break
                    if (rule, func) == last_items:
                        self.throw_error(
                            ErrorType.SYNTAX,
                            f"Invalid character: '{self.char}'",
                            char=self.char
                        )
        except LPV_Exception as e:
            raise e
        except Exception as e:
            if crash_handler is print:
                crash_handler(traceback.format_exc())
            else:
                crash_handler(e)
            self.throw_error(
                ErrorType.CRASH,
                "Uh oh. Looks like the lexer got an Python Error."
            )
        return TokenTree(self.source, *tree)


import textwrap

class NodePrettier(object):
    def __init__(self, parent, **childs):
        self.parent = parent
        self.childs = childs
    def __repr__(self):
        return build_tree(self)

def build_tree(node: NodePrettier):
    string = f"[:{node.parent}:]"
    items = tuple(node.childs.items())
    for cname, cvalue in items:
        head = f"[{cname}]"
        string += f"\n    "+head
        string += "\n"+textwrap.indent(
            cvalue.__repr__(),
            prefix="    "*2
        )
        string += "\n    [end]"
    return string+"\n[:end:]"


# MIT License

# Copyright (c) 2021 xp

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from LPV.src.error import ErrorType, LPV_Exception
from LPV.src.token import Token, TokenTree, TokenType
from typing import Callable, Iterable, Union
import traceback

class LPV_Parser:
    initiliaze = False
    def __init__(self):
        if not "start" in self.__dict__:
            raise TypeError("no self.start found")
        if not isinstance(self.start, Iterable):
            raise TypeError(
                "self.start must be iterable"
            )
        if len(self.start) < 2:
            raise TypeError(
                "self.start must have two element: function/function_name and arguments"
            )
        self.method, self.args = self.get_method(self.start[0]), self.start[1]
        self.initiliaze = True
        
    def get_method(self, func: Union[str, Callable]) -> Callable:
        if isinstance(func, str):
            if func.startswith("parse_"):
                name = func
            else:
                name = "parse_"+func
            m = getattr(self, name)
            if not isinstance(m, Callable):
                raise TypeError(f"{name} must be callable, not {type(m).__name__}")
            return m
        if not func.__name__.startswith("parse_"):
            raise TypeError(f"{func.__name__} name must start with parse_")
        return func
    
    def init(self, token_tree: TokenTree):
        self.tree, self.source = token_tree, token_tree.source
        self.token = None
        self.line, self.col = -1, -1
        self.next_token()
    
    def get_lc(self):
        return self.line, self.col
    
    def throw_error(self, error, msg, pointer_width:int=1, **kwargs):
        raise LPV_Exception(
            error=error, msg=msg,
            line=self.line, col=self.col,
            source=self.source, pointer_width=pointer_width,
            when="Parser", **kwargs
        )
    
    def next_token(self):
        old_t = self.token
        self.token = self.tree.next_token()
        self.line, self.col = self.tree.line, self.tree.col
        return old_t
    
    def is_eof(self):
        return self.token is None
    
    def eat(self, type_: Union[TokenType, Iterable]=None):
        if type_ is None or self.check_type(type_):
            return self.next_token()
        msg = "Unexpected "+("EOF" if self.is_eof() else f"'{self.token.value}'")
        self.throw_error(
            ErrorType.SYNTAX, msg+", expected "+(
                type_.name if isinstance(type_, TokenType) else " or ".join(type_)
            ),
        )
    
    def peek(self, step:int=1):
        return self.tree.peek(step)
        
    def check_type(self, type_: Union[TokenType, Iterable[TokenType]], token=False):
        if not isinstance(type_, (TokenType, Iterable)):
            raise TypeError("type_ must be TokenType or Iterable")
        if isinstance(type_, TokenType):
            if token is not False:
                return token == type_
            return self.token == type_
        if token is not False:
            for t in type_:
                if self.check_type(t, token=token):
                    return True
            return False
        for t in type_:
            if self.check_type(t):
                return True
        return False
    
    def eats(self, *types) -> list[Token]:
        tkns = []
        for t in types:
            tkns.append(self.eat(t))
        return tkns
    
    def match(self, *types):
        for i, t in enumerate(types):
            if not self.check_type(t, self.peek(i)):
                return False
        return True
    
    def parse(self, tree: TokenTree, crash_handler:Callable=print):
        if self.initiliaze is False:
            raise TypeError(
                f"{type(self).__name__} is not fully initiliaze. Forgot super().__init__() in the __init__?"
            )
        self.init(tree)
        try:
            r = self.method(*self.args)
        except LPV_Exception as e:
            raise e
        except Exception as e:
            if crash_handler is print:
                crash_handler(traceback.format_exc())
            else:
                crash_handler(e)
            self.throw_error(
                ErrorType.CRASH,
                "Uh oh. Looks like the parser got an Python Error."
            )
        else:
            return r


# MIT License

# Copyright (c) 2021 xp

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from typing import Iterable

class TokenType:
    
    def __init__(self, name: str) -> None:
        self.name = name
    def __repr__(self):
        return self.name
    def __eq__(self, o: str):
        return self.name == o

class Token:
    
    def __init__(self, type_: TokenType, value) -> None:
        if not isinstance(type_, TokenType):
            raise TypeError(
                f"type_ argument must be TokenType not {type(type_).__name__}"
            )
        self.type, self.value = type_, value
        self.line, self.col = None, None
    def set_lc(self, line:int, col:int):
        self.line, self.col = line, col
    def __repr__(self) -> str:
        return "Token({type}, '{value}')".format(
            type=self.type,value=self.value
        )
    def __eq__(self, o: object) -> bool:
        return (self.type == o) if isinstance(o, TokenType) else (self.value == o)

def generate_tokens(name: str, tokens: Iterable[str], build:bool=False):
    if not name.isidentifier():
        raise TypeError(
                f"Invalid class name: {name}"
            )
    for t in tokens:
        if not t.isidentifier():
            raise TypeError(
                f"Invalid token name: {t}"
            )
    code = """\
class {name}:
{attributes}\
""".format(
            name=name,
            attributes="\n".join(
                (f"    {t} = TokenType('{t}')" for t in tokens)
            )
        )
    if build is True:
        return code
    exec(code)
    return eval(name)

class TokenTree:
    
    def __init__(self, source: str, *tokens: Token) -> None:
        self.source = source
        self.tokens, self.len_t = tokens, len(tokens)
        self.pos, self.token = -1, None
        self.line, self.col = 0, 0
    
    def __getitem__(self, index):
        return self.tokens[index]
    
    def next_token(self):
        self.pos += 1
        if self.pos >= self.len_t:
            self.token = None
            return self.token
        else:
            self.token = self.tokens[self.pos]
            self.line, self.col = self.token.line, self.token.col
            return self.token
    
    def peek(self, step:int=1):
        p = self.pos+step
        if p >= self.len_t:
            return None
        return self.tokens[p]
    
    def get(self, o: object, default=None):
        for t in self.tokens:
            if t == o:
                return t
        return default
    
    def __repr__(self):
        string = "TokenTree:\n"
        space = " "*len(string[:-1])
        for token in self.tokens:
            head = f"{token.type}:"
            content = f"{' '*len(head)}\"{token.value}\""
            string += space+head+"\n"+space+content+"\n"
        return string